from utils import standardize, sampled_sphere
import numpy as np

def AI_IRW(X, AI=True, robust=False, n_dirs=None, X_test=None, random_state=None):
    """ Compute the score of the (Affine-invariant-) integrated rank 
        weighted depth of X_test w.r.t. X
    Parameters
    ----------
    X: Array of shape (n_samples, n_features)
            The training set.
    AI: bool
        if True, the affine-invariant version of irw is computed. 
        If False, the original irw is computed.
    robust: bool, default=False
        if robust is true, the MCD estimator of the covariance matrix
        is performed.
    n_dirs: int | None
        The number of random directions needed to approximate 
        the integral over the unit sphere.
        If None, n_dirs is set as 100* n_features.
    X_test: Array of shape (n_samples_test, n_features)
        The testing set. 
        If None, return the score of the training sample.
    random_state: int | None
        The random state.
        
    Returns
    -------
    ai_irw_score: Array
        Depth score of each element in X_test.
        If X_test is None, return the score of the training sample.
    """
    

    #Setting seed:
    if random_state is None:
        random_state = 0

    np.random.seed(random_state)




    if X_test is None:

        if AI and len(X)<10000:
            X_reduced, _ = standardize(X, robust)
        else:
            X_reduced = X.copy()

        n_samples, n_features = X_reduced.shape 
        
        #Setting the number of directions to 100 times the number of features as in the paper.
        if n_dirs is None:
            n_dirs = n_features * 100

        #Simulated random directions on the unit sphere. 
        U = sampled_sphere(n_dirs, n_features)

        sequence = np.arange(1, n_samples + 1)
        depth = np.zeros((n_samples, n_dirs))
        
        proj = np.matmul(X_reduced, U.T)
        rank_matrix = np.matrix.argsort(proj, axis=0)
        
        for k in range(n_dirs):
            depth[rank_matrix[:, k], k] = sequence  
        
        depth =  depth / (n_samples * 1.)         

        ai_irw_score = np.mean(np.minimum(depth, 1 - depth), axis = 1)

    else:

        if AI and len(X)<10000:
            X_reduced, Sigma_inv_square = standardize(X, robust)
            X_test_reduced = X_test@Sigma_inv_square
        else:
            X_reduced = X.copy()
            X_test_reduced = X_test.copy()

        n_samples, n_features = X_reduced.shape
        n_samples_test, _ = X_test_reduced.shape 

        #Setting the number of directions to 100 times the number of features as in the paper.
        if n_dirs is None:
            n_dirs = n_features * 100

        #Simulated random directions on the unit sphere. 
        U = sampled_sphere(n_dirs, n_features)

        proj = np.matmul(X_reduced, U.T)
        proj_test = np.matmul(X_test_reduced, U.T)

        sequence = np.arange(1, n_samples_test+1)
        depth = np.zeros((n_samples_test, n_dirs))
        temp = np.zeros((n_samples_test, n_dirs))

        proj.sort(axis=0)
        for k in range(n_dirs):
            depth[:,k] = np.searchsorted(a=proj[:,k],v=proj_test[:,k], side='left')

        depth /=   n_samples * 1. 

        ai_irw_score = np.mean(np.minimum(depth, 1 - depth), axis=1)

    return ai_irw_score



import numpy as np
from codecarbon import OfflineEmissionsTracker
from sklearn.metrics import *
import time 
import pandas as pd
from sklearn.metrics import roc_auc_score, r2_score, accuracy_score, f1_score, balanced_accuracy_score, matthews_corrcoef, mean_absolute_error, mean_squared_error

def predict_model(model, x, project_type, threshold=0.5):
    """
    the model class is only assumed to have a predict_proba method for classification problems.
    
    :project_type: can be either classification_b classification_m or regression
    """
    probas, x_test = 0,0
    if project_type == "classification_b":
        preds = (model.predict_proba(x) >= threshold)[:,1].astype(int)
    elif project_type == "classification_m":
        preds = np.asarray([np.argmax(probas[i,:]) for i in range(len(x_test))])
    return preds

def check_and_transform(Y):
    if isinstance(Y, np.ndarray):
        pass
    elif isinstance(Y, pd.DataFrame):
        Y = Y.to_numpy()
    elif isinstance(Y, list):
        Y = np.asarray(Y)
    return Y


#Set of admissible parameters for custom metric definition

#y_pred: Predicted target classes for x_test.
#probs: Predicted probabilities for each class for x_test. 
#y_test: Ground truth target classes for x_test.
#threshold: The model's threshold.
#model: The model on which to evaluate the metric.
#x_test: the testing set.
#x_train: the training set.
#y_train: Ground truth target for x_train.
#protected_attribute: protected attribute (dict): {
#        'column_name': 'column_name',
#        'type': 'cat/float',
#        'value': 0 if 'cat', [min, max] if 'float',
#        'name_prot_attr': 'understandable name'}

#################METRICS CLASSIFICATION#################

def roc_curve(y_pred, y_test):
    fp = np.sum((y_pred == 1) & (y_test == 0))
    tp = np.sum((y_pred == 1) & (y_test == 1))
    fn = np.sum((y_pred == 0) & (y_test == 1))
    tn = np.sum((y_pred == 0) & (y_test == 0))
    fpr = fp / (fp + tn)
    tpr = tp / (tp + fn)
    tnr = tn / (tn + fp)
    fnr = fn / (tp + fn)
    return fpr, tpr, fnr, tnr

def fpr(y_pred, y_test):
    """
    The False Positive Rate corresponds to the proportions of wrong predictions among the samples that are true negative. 
    For a credit scoring application, it means according a loan to an individual which would have not pay back.
    The lower, the better.

    """
    fp = np.sum((y_pred == 1) & (y_test == 0))
    tp = np.sum((y_pred == 1) & (y_test == 1))
    fn = np.sum((y_pred == 0) & (y_test == 1))
    tn = np.sum((y_pred == 0) & (y_test == 0))
    fpr = fp / (fp + tn)
    tpr = tp / (tp + fn)
    tnr = tn / (tn + fp)
    fnr = fn / (tp + fn)
    return fpr

def tpr(y_pred, y_test):
    """
    The True Positive Rate corresponds to the proportions of right predictions among the samples that are true positive.
    For a credit scoring application, it means according a loan to an individual which would have pay back. 
    It is also called the recall or the sensitivity of the model.
    The higher, the better.

    """
    fp = np.sum((y_pred == 1) & (y_test == 0))
    tp = np.sum((y_pred == 1) & (y_test == 1))
    fn = np.sum((y_pred == 0) & (y_test == 1))
    tn = np.sum((y_pred == 0) & (y_test == 0))
    fpr = fp / (fp + tn)
    tpr = tp / (tp + fn)
    tnr = tn / (tn + fp)
    fnr = fn / (tp + fn)
    return tpr

def fnr(y_pred, y_test):
    """
    The False Negative Rate corresponds to the proportions of wrong predictions among the samples that are true positive. 
    For a credit scoring application, it means denying a loan to an individual which would have pay back.
    The lower, the better.

    """
    fp = np.sum((y_pred == 1) & (y_test == 0))
    tp = np.sum((y_pred == 1) & (y_test == 1))
    fn = np.sum((y_pred == 0) & (y_test == 1))
    tn = np.sum((y_pred == 0) & (y_test == 0))
    fpr = fp / (fp + tn)
    tpr = tp / (tp + fn)
    tnr = tn / (tn + fp)
    fnr = fn / (tp + fn)
    return fnr

def tnr(y_pred, y_test):
    """
    The True Negative Rate corresponds to the proportions of wrong predictions among the samples that are true positive. 
    For a credit scoring application, it means denying a loan to an individual which would have not pay back.
    The higher, the better.

    """
    fp = np.sum((y_pred == 1) & (y_test == 0))
    tp = np.sum((y_pred == 1) & (y_test == 1))
    fn = np.sum((y_pred == 0) & (y_test == 1))
    tn = np.sum((y_pred == 0) & (y_test == 0))
    fpr = fp / (fp + tn)
    tpr = tp / (tp + fn)
    tnr = tn / (tn + fp)
    fnr = fn / (tp + fn)
    return tnr

def compute_Matthews_corr_coef_b(y_pred, y_test):
    """
    The function computes the MMC score (between -1 and 1) that summarize how good the classification is. 
    The Matthews correlation coefficient (MCC) is a more reliable statistical rate which produces a high score only if the prediction obtained good results in all of the four confusion matrix categories (true positives, false negatives, true negatives, and false positives), proportionally both to the size of positive elements and the size of negative elements in the dataset. Reference: https://bmcgenomics.biomedcentral.com/articles/10.1186/s12864-019-6413-7. In other words, MCC is the only binary classification rate that generates a high score only if the binary predictor was able to correctly predict the majority of positive data instances and the majority of negative data instances. If a confusion matrix threshold is at disposal, the litterature recommends the usage of the Matthews correlation coefficient over F1 score, and accuracy.
    
    Drawbacks: 
    
    :y_pred: Predicted target classes.
    :y_test: Ground truth target classes.
    """
    return matthews_corrcoef(y_test, y_pred)


#def compute_Matthews_corr_coef_m(y_pred, y_test):
#    """
#    The function computes the MMC score (between -1 and 1) that summarize how good the classification is. 
#    
#    Advantages: Statistical measures like accuracy or F1 score can dangerously show overoptimistic inflated results, especially on imbalanced datasets. The Matthews correlation coefficient (MCC), instead, is a more reliable statistical rate which produces a high score only if the prediction obtained good results in all of the four confusion matrix categories (true positives, false negatives, true negatives, and false positives), proportionally both to the size of positive elements and the size of negative elements in the dataset. Reference: https://bmcgenomics.biomedcentral.com/articles/10.1186/s12864-019-6413-7. In other words, MCC is the only binary classification rate that generates a high score only if the binary predictor was able to correctly predict the majority of positive data instances and the majority of negative data instances. If a confusion matrix threshold is at disposal, the litterature recommends the usage of the Matthews correlation coefficient over F1 score, and accuracy.
    
#    Drawbacks: 
    
#    :y_pred: Predicted target classes.
#    :y_test: Ground truth target classes.
#    """
#    return matthews_corrcoef(y_test, y_pred)


def compute_expected_calibration_error_b(model, y_test, x_test,probs):
    """
    The Expected Calibration Error (ECE) measures how well a model's probabilistic predictions reflect the risk they involve.

    :probs: Predicted probabilities for each class.
    :y_test: Ground truth target classes.
    """
    labels = np.unique(y_test)
    if len(labels) > 2:
        raise ValueError(
            "Only binary classification is supported. Provided labels %s." % labels
        )
    y_prob = probs[:,1]
    n_bins = np.sqrt(len(y_prob)).astype(int)
    quantiles = np.linspace(0, 1, n_bins + 1)
    bins = np.percentile(y_prob, quantiles * 100)
    bins[-1] = bins[-1] + 1e-8
    binids = np.digitize(y_prob, bins) - 1
    bin_sums = np.bincount(binids, weights=y_prob, minlength=len(bins))
    bin_true = np.bincount(binids, weights=y_test, minlength=len(bins))
    bin_total = np.bincount(binids, minlength=len(bins))
    nonzero = bin_total != 0
    prob_true = bin_true[nonzero] / bin_total[nonzero]
    prob_pred = bin_sums[nonzero] / bin_total[nonzero]
    exp_calib_error = (bin_total[nonzero] * np.abs(prob_true - prob_pred)).sum()/len(y_test)
    return exp_calib_error

def compute_roc_auc_score(y_pred, y_test):
    """
    The AUC (Area Under the Curve) measures to which extent a model has learn to discriminate between positive and negative samples, 
    at the level of the predicted scores. 
    Taking a pair of random positive sample P and random negative sample N, it is equal to the probability that the score of P is higher
    than the score of N. 
    Notice that this metric is independent of the choice of the threshold.
    The higher, the better. 

    """
    return roc_auc_score(y_test, y_pred)


def compute_accuracy(y_pred, y_test):
    """
    The accuracy corresponds to the proportion of good predictions. 

    """
    return accuracy_score(y_test, y_pred)


def compute_balanced_accuracy_adjusted(y_pred, y_test):
    """
    The function computes the balanced accuracy. It is defined as the average of recall obtained on each class. Random performance  scores 0. Perfect performance has a score of 1.
    
    Reference: https://scikit-learn.org/stable/modules/generated/sklearn.metrics.balanced_accuracy_score.html
    
    :y_pred: Predicted target classes.
    :y_test: Ground truth target classes.
    """
    return balanced_accuracy_score(y_test, y_pred, adjusted = True)


def compute_f1_score_b(y_pred, y_test):
    """
    The F1 score is a measure that aggregates the precision (the proportion of good predictions among the samples that are predicted positive) 
    and the recall (which is the True Positive Rate) of a model. 
    A F1 score of X means that for one model error, there are 2(1-X)/X true positive. In the particular case where X=1/2, it means that to one misclassified
    sample corresponds two true positive samples.
    The higher, the better.

    """
    return f1_score(y_test, y_pred)


#def compute_f1_score_m(y_pred, y_test):
#    """
#    The function computes the f1 score for binary classifiers. It Calculate metrics globally by counting the total true positives, false negatives and false positives.
    
#    Drawbacks: F1 score, although popular, can generate misleading results on imbalanced datasets, because it fails to consider the ratio between positive and negative elements.
    
#    Reference: https://scikit-learn.org/stable/modules/generated/sklearn.metrics.f1_score.html
    
#    :y_pred: Predicted target classes.
#    :y_test: Ground truth target classes.
#    """
#    return f1_score(y_test, y_pred, average='micro')



    
def compute_emissionCO2_classification(model, x_test, y_test):
    """
    The function computes the CO2 emissions that are produced by model training. It uses 
    the tracker from the codecarbon project. Reference: https://mlco2.github.io/codecarbon/

    :model: the model on which to evaluate the metric.
    :x_test: the testing set.
    :y_test: Ground truth target classes.
    """
    tracker = OfflineEmissionsTracker(country_iso_code="FRA", log_level='critical', save_to_file = False)
    emissions_ = 0
    tracker.start()  
    try:
        model.predict_proba(x_test)                           
    finally:
        emissions_  = tracker.stop()
    return emissions_*1e9

    
def compute_inferenceTime_classification(model, x_test):
    """
    The function computes the inference time which is averaged on the test set. Note that depending on the design of the network architecture, the architecture complexity does not necessarily reflect the computational requirements for performing network inference (e.g., MobileNet has more parameters than SqueezeNet but has lower computational requirements for network inference). Reference: https://arxiv.org/pdf/1806.05512.pdf

    :model: the model on which to evaluate the metric.
    :x_test: the testing set.

    """
    t1 = time.time()
    preds = model.predict_proba(x_test)
    t2 = time.time()
    time_predict_ = float((t2-t1))/len(x_test)
    return time_predict_*1e6



def compute_NetTrustScore_b(probs, y_pred, y_test):
    """
    The function computes the NetTrustScore, a global scalar metric summarizing overall trustworthiness of a model. A model that has higher confidence when predicting correctly will receive a higher marginal trust score for that prediction. A model that has higher confidence when predicting incorrectly will receive a lower marginal trust for that prediction. Scores are then agregated across all predictions. Reference: https://arxiv.org/pdf/2009.05835.pdf
    
    :probs: Predicted probabilities for each class.
    :y_pred: Predicted target classes.
    :y_test: Ground truth target classes.
    """
    n = len(y_test)
    alpha=1
    beta=1
    if len(y_test.shape)>1:
        y_test = y_test.flatten()
    scores = []
    for num_class in set(y_test):
        indices = np.where(y_test == num_class)[0]
        len_class = len(indices)  
        correct_preds_indices = [i for (i, elt) in enumerate(y_pred) if elt == num_class and y_test[i]==num_class]
        false_preds_indices = [i for (i, elt) in enumerate(y_pred) if elt != num_class and y_test[i]==num_class]
        correct_preds = probs[correct_preds_indices, num_class]**alpha
        false_preds = (1- probs[false_preds_indices, 1-num_class])**beta   
        QA_trust = np.concatenate([correct_preds, false_preds], axis=0).reshape([-1,1])    
        trust_score = (1/len(QA_trust))*sum(QA_trust)
        proba_to_get_class = (1/n)*len_class
        scores.append(proba_to_get_class*trust_score)
    return float(sum(scores))


def fairness_DI_demographic_parity_b(y_pred, x_test, y_test, protected_attribute):
    """
    The disparate impact computed for the fairness demographic parity metric.
    
    :y_pred: Predicted target classes.
    :x_test: the testing set.
    :y_test: Ground truth target classes.
    :protected_attribute: {
        'column_name': 'column_name',
        'type': 'cat/float',
        'value': 0 if 'cat', [min, max] if 'float',
        'name_prot_attr': 'understandable name'
    }
    """
    # subgroup slices
    c_name = protected_attribute['column_name']
    if protected_attribute['type'] == 'cat':
        protected_mask = x_test[c_name] == protected_attribute['value']
    if protected_attribute['type'] == 'float':
        c_min, c_max = protected_attribute['value']
        protected_mask = ( x_test[c_name] >= c_min ) & ( x_test[c_name] <= c_max )
    # accuracy for each subgroup
    protected_accuracy = (y_pred[protected_mask] == y_test[protected_mask]).mean()
    favored_accuracy = (y_pred[~protected_mask] == y_test[~protected_mask]).mean()
    # sanity checks
    if protected_accuracy == 0:
        if favored_accuracy == 0:
            res = 1.
        else:
            res = np.Infinity # should print warning
    else: 
        res = favored_accuracy / protected_accuracy 
    return res



#################METRICS REGRESSION#################

def compute_inferenceTime_regression(model, x_test):
    """
    The function computes the inference time which is averaged on the test set. Note that depending on the design of the network architecture, the architecture complexity does not necessarily reflect the computational requirements for performing network inference (e.g., MobileNet has more parameters than SqueezeNet but has lower computational requirements for network inference). Reference: https://arxiv.org/pdf/1806.05512.pdf

    :model: the model on which to evaluate the metric.
    :x_test: the testing set.

    """
    t1 = time.time()
    preds = model.predict(x_test)
    t2 = time.time()
    time_predict_ = float((t2-t1))/len(x_test)
    return time_predict_

def compute_emissionCO2_regression(model, x_test, y_test):
    """
    The function computes the CO2 emissions that are produced by model training. It uses 
    the tracker from the codecarbon project. Reference: https://mlco2.github.io/codecarbon/

    :model: the model on which to evaluate the metric.
    :x_test: the testing set.
    :y_test: Ground truth target classes.
    """
    tracker = OfflineEmissionsTracker(country_iso_code="FRA", log_level='warning', save_to_file = False)
    emissions_ = 0
    tracker.start()  
    try:
        model.predict(x_test)                           
    finally:
        emissions_  = tracker.stop()
    return emissions_


def compute_r2(y_pred, y_test):
    """
    The function computes the R-squared for regression models. R-squared explains to what extent the variance of independant input variable explains the variance of the predicted output variable.
    
    Reference: https://scikit-learn.org/stable/modules/generated/sklearn.metrics.r2_score.html
    
    :y_pred: Predicted target classes.
    :y_test: Ground truth target classes.
    """
    return r2_score(y_test, y_pred)

def compute_mean_absolute_error(y_pred, y_test):
    """
    The function computes the mean absolute error (MAE), also called the L1 loss for regression models. 
    Advantage: less affected by outliers than the L2 error i.e. more robust. Note that if the use case does require to pay attention to the outliers, it might be an drawback rather than an advantage and MSE might be a better choice.
    Drawbacks: as a loss function, it is not differentiable (i.e. it might present difficulties in the gradient computations during backpropagation). It can be unstable (the solution i.e. the decision boundary is NOT changing smoothly with small data changes).
    
    Reference: https://scikit-learn.org/stable/modules/generated/sklearn.metrics.mean_absolute_error.html
    
    :y_pred: Predicted target classes.
    :y_test: Ground truth target classes.
    """
    return mean_absolute_error(y_test, y_pred)

def compute_mean_squared_error(y_pred, y_test):
    """
    The function computes the mean squared error (MSE), also called the L2 loss for regression models. 
    Advantages: the MSE is stable (the solution i.e. the decision boundary is changing smoothly with small data changes) and differentiable.
    Drawbacks: the MSE is more affected by outliers than the L1 error i.e. less robust. Note that if the use case does require to pay attention to the outliers, it might be an advantage rather than a drawback and MSE might be a better choice.
    Reference: https://scikit-learn.org/stable/modules/generated/sklearn.metrics.mean_squared_error.html
    
    :y_pred: Predicted target classes.
    :y_test: Ground truth target classes.
    """
    return mean_squared_error(y_test, y_pred)


metrics = [
    {'metric_name' : 'Accuracy', 'metric_function' : compute_accuracy, 'metric_type' : 'classif', 'h_is_b' : True },
    {'metric_name' : 'True positives', 'metric_function' : tpr, 'metric_type' : 'classif', 'h_is_b' : True },
    {'metric_name' : 'False positives', 'metric_function' : fpr, 'metric_type' : 'classif', 'h_is_b' : False },
    {'metric_name' : 'False negatives', 'metric_function' : fnr, 'metric_type' : 'classif', 'h_is_b' : False },
    {'metric_name' : 'True negatives', 'metric_function' : tnr, 'metric_type' : 'classif', 'h_is_b' : True },
    {'metric_name' : 'ROC AUC', 'metric_function' : compute_roc_auc_score, 'metric_type' : 'classif', 'h_is_b' : True },
    {'metric_name' : 'F1 score', 'metric_function' : compute_f1_score_b, 'metric_type' : 'classif', 'h_is_b' : True },
    {'metric_name' : 'CO2 Emission (reg)', 'metric_function' : compute_emissionCO2_regression, 'metric_type' : 'reg', 'h_is_b' : False },
    {'metric_name' : 'Inference time (reg)', 'metric_function' : compute_inferenceTime_regression, 'metric_type' : 'reg', 'h_is_b' : False },
    {'metric_name' : 'R2 score', 'metric_function' : compute_r2, 'metric_type' : 'reg', 'h_is_b' : True },
    {'metric_name' : 'Mean squared error', 'metric_function' : compute_mean_squared_error, 'metric_type' : 'reg', 'h_is_b' : False },
    {'metric_name' : 'Mean absolute error', 'metric_function' : compute_mean_absolute_error, 'metric_type' : 'reg', 'h_is_b' : False },
#    {'metric_name' : 'F1 score micro', 'metric_function' : compute_f1_score_m, 'metric_type' : 'classif', 'h_is_b' : True },
    {'metric_name' : 'CO2 Emissions', 'metric_function' : compute_emissionCO2_classification, 'metric_type' : 'classif', 'h_is_b' : False },
    {'metric_name' : 'Inference time', 'metric_function' : compute_inferenceTime_classification, 'metric_type' : 'classif', 'h_is_b' : False },
    {'metric_name' : 'Balanced accuracy adjusted', 'metric_function' : compute_balanced_accuracy_adjusted, 'metric_type' : 'classif', 'h_is_b' : True },
    {'metric_name' : 'Trust Score', 'metric_function' : compute_NetTrustScore_b, 'metric_type' : 'classif', 'h_is_b' : True },
    {'metric_name' : 'Expected calibration error', 'metric_function' : compute_expected_calibration_error_b, 'metric_type' : 'classif', 'h_is_b' : True },
    {'metric_name' : 'Matthews correlation', 'metric_function' : compute_Matthews_corr_coef_b, 'metric_type' : 'classif', 'h_is_b' : True },
#    {'metric_name' : 'Matthews correlation coeffficient m', 'metric_function' : compute_Matthews_corr_coef_m, 'metric_type' : 'classif', 'h_is_b' : True },
    {'metric_name' : 'Fairness DI demographic parity b', 'metric_function' : fairness_DI_demographic_parity_b, 'metric_type' : 'classif', 'h_is_b' : True },
#    {'metric_name' : 'Inference Time', 'metric_function' : compute_inferenceTime_classification, 'metric_type' : 'classif', 'h_is_b' : False }
]

def add_doc(m) :
    m['description'] = m['metric_function'].__doc__ 
    return m
d_metrics = [add_doc(m) for m in metrics]



from ast import arg
from gc import get_stats
import inspect
import io
import json
import marshal
import math
from multiprocessing.sharedctypes import Value
import types
from unittest.mock import Mock
import requests
import pandas as pd
import pprint
import pickle
from copy import deepcopy

from althiqa.default_metrics import metrics as d_metrics

def add_doc(m) :
    m['description'] = m['metric_function'].__doc__ 
    return m
d_metrics = [add_doc(m) for m in d_metrics]

from codecarbon import OfflineEmissionsTracker

from sklearn.metrics import roc_auc_score, r2_score, accuracy_score, f1_score, balanced_accuracy_score, matthews_corrcoef

from althiqa.aiirw import AI_IRW

import copy



import numpy as np
from sklearn.metrics import roc_auc_score, r2_score, accuracy_score, f1_score, balanced_accuracy_score, matthews_corrcoef
import math
import pickle
import io
import time


MAX_MB_UPLOAD = 1

def check_and_transform(Y):
    if isinstance(Y, np.ndarray):
        pass
    elif isinstance(Y, pd.DataFrame):
        Y = Y.to_numpy()[:,0]
    elif isinstance(Y, list):
        Y = np.asarray(Y)
    return Y

def compute_aiirw_depth(x_train, y_train, x_test, y_test):
    """Compute the aiirw depths of data points in x_test based on how they are far away from the
    in distribution constituted of x_train and x_test."""
    
    y_train = check_and_transform(y_train)
    y_test = check_and_transform(y_test)
    #print(np.concatenate([y_train,y_test]))
    unique_classes = set(np.concatenate([y_train,y_test]))
    scores_aiirw = []
    depth = [0]*len(y_test)
    x_test_depth = copy.deepcopy(x_test)
    for lab in unique_classes:
    
        indices_test = np.where(y_test == lab)[0]
        indices_train = np.where(y_train == lab)[0]

        x_train_lab = x_train.iloc[indices_train]
        x_test_lab = x_test.iloc[indices_test]
        #print(np.concatenate([x_train_lab, x_test_lab]))
        score_aiirw = AI_IRW(np.concatenate([x_train_lab, x_test_lab]))
        score_aiirw_test = score_aiirw[len(x_train_lab):]
        
        index_test  = x_test_depth.index[indices_test] 
        x_test_depth.loc[index_test, 'depth'] = score_aiirw_test

    return x_test_depth["depth"]

def get_rank(depth):
    """Compute the rank of each sample in x_test based on how far they are in the distribution. Smaller rank means
    closer to the center of the distribution"""
    return len(depth) - np.argsort(np.argsort(depth))


def compute_acc_vs_depth(depth_ranks, y_pred, y_test):
    """
    Function that compute the robustness curve every 1% of depth.
    
    Back computation.
    
    depth_ranks: array of int, for each test input, the rank for the depth (high depth means in-distribution).
    y_pred: array, the predictions of the model.
    y_test: array, the ground truth
    
    Outputs: prop, acc
    
    prop: list of the proportions (in pourcentage), 
          of size 100 if len(depth_ranks)>=100, otherwise of size len(y_test)
    acc: list of the accuracies corresponding to the prop list
    """
    
    # order y_pred and y_test according to the depth
    depth_ranks_ = np.array(depth_ranks) - 1
    y_pred_ = y_pred[depth_ranks_]
    y_test_ = y_test[depth_ranks_]
    
    s = 0  # the cumulated number of good predictions

    acc = []
    if len(depth_ranks) >= 100:
        prop = np.arange(100) + 1
        n = int((1/100)*len(depth_ranks)) # the amount of samples in 1%
        for i in range(len(depth_ranks)):
            s += int(y_pred_[i] == y_test_[i])
            if (i+1) % n == 0:
                acc.append(s / (i+1))
            if i == len(depth_ranks) - 1:
                if len(acc) == 99:
                    acc.append(s / (i+1))
    else:
        prop = []
        for i in range(len(depth_ranks)):
            s += int(y_pred_[i] == y_test_[i])
            acc.append(s / (i+1))
            prop.append(round((i+1)*100/len(depth_ranks),1))
    prop_and_acc = [[float(p),float(a)] for p,a in zip(prop,acc)]
    return prop_and_acc

        
class MockModel():
    p = 0.5

    def predict(self,df) :
        return [self.p]*len(df)



def compute_rescaled_scores(dict_score_models, h_is_b):
    """
    Compute rescaled scores from raw scores.
    
    Back computation: Use it each time a model or a metric is pushed.
    
    dict_score_models: dict of the form
    dict_score_models_test = {
        'model_1': 0.,
        'model_2': 10.,
        'model_3': 100.,
        'model_4': 50.
    }
    
    h_is_b: Boolean associated to the metric ("higher is better")
    
    Output: dict of the form
    {
    'model_1': 0.2, 
    'model_2': 0.1, 
    'model_3': 1., 
    'model_4': 0.5
    }
    """
    
    scores = np.array(list(dict_score_models.values()))
    min_, max_ = min(scores), max(scores)

    output = {}
    if len(dict_score_models) >= 3:
        if min_ != max_:
            output = {model: (score - min_) / (max_ - min_) for model, score in dict_score_models.items()}
            if h_is_b is False:
                output = {model: 1 - score for model, score in output.items()} 
            output = {model: 0.5 + (score-0.5)/(0.5/0.3) for model, score in output.items()}
        else:
            output = {model: 0.6 for model in output.keys()}
    else:
        output = {model: None for model in output.keys()} 
    return output

    

class Project():
    def __init__(self, session, pk, project_name :str, project_type, x_train, y_train, x_test, y_test, models=[],metrics=[],depth_metrics=[],scores=[],reports=[]) -> None:
        if project_type.lower() not in ['reg', 'classif']:
            raise ValueError("project_type should be 'reg' or 'classif', {} is not a valid project type.".format(project_type)) 
        
        self.project_type = project_type.lower()
        self.pk = pk
        self.project_name = project_name
        self.x_train = x_train
        self.y_train = y_train
        self.x_test = x_test
        self.y_test = y_test
        self._models = models
        self._model_objects = None
        self._metrics = metrics
        self._depth_metrics = depth_metrics
        self._metrics_objects = None
        self._scores = scores
        self.reports = reports
        self.session = session


    def list_models(self):
        md = deepcopy(self._models)
        print([mod.pop("model_file", None).pop("y_pred") for mod in md])

    def get_model(self, id) :
            #print(id)
            try :
                m=  [mod for mod in self._models if mod['model_name'] == id ][0]
            except :
                raise NameError("No model with the name {name_or_id}. List models with the Project.list_models() method.")
            try :
                r = requests.get(m['model_file'])
                b= io.BytesIO()
                b.write(r.content)
                b.seek(0)
                return pickle.load(b)
            except :
                raise Exception("Error while loading the model")

    def push_model(self, model_name, model, threshold=None):
        """
        Push a ML Model to althiqa's platform. 

        The model should be an object with a predict() method if it is a regression project.
        The model should be an object with a predict_proba() model if it is a classification project, 
        and a threshold between 0 and 1 should be given as argument. 
        """
        if self.project_type == 'reg' : 
            predict  = getattr(model, "predict", None)
            if not callable(predict):
                raise Exception(f'{model} does not have a predict method')
            y_pred = pd.DataFrame(model.predict(self.y_test))
            scores = {}
            depth_scores = []
            for metric in self._metrics :
                score = self.compute_metric(metric, model, threshold)
                if score is not None: 
                    scores[metric['metric_name']] = score
        
            for metric in self._depth_metrics :
                depth_scores.append({'depth_metric_name' : metric['depth_metric_name'] , 'score' : self.compute_depth_metric(metric, model, threshold) })
        
        if self.project_type == 'classif' : 
            if threshold is None or float(threshold)<0 or float(threshold)>1 : raise ValueError("This is a classification project, please enter a threshold value between 0 and 1.")
            predict  = getattr(model, "predict_proba", None)
            if not callable(predict):
                raise Exception(f'{model} does not have a predict_proba method')
            y_pred = pd.DataFrame(model.predict_proba(self.x_test))
            y_pred = (y_pred>float(threshold)).iloc[:, 1].astype(int)
            scores = {}
            for metric in self._metrics :
                score = self.compute_metric(metric, model, threshold)
                if score is not None: 
                    scores[metric['metric_name']] = score
            depth_scores = []
            #(self._depth_metrics)
            for metric in self._depth_metrics :
                depth_scores.append({'depth_metric_name' : metric['depth_metric_name'] , 'score' : self.compute_depth_metric(metric, model, threshold) })
        

        headers = {}
        headers["Accept"] = "application/json"
        headers["Authorization"] = "Bearer {token}".format(token  = self.session.token)
        buffer_y_pred = io.BytesIO()
        y_pred.to_pickle(buffer_y_pred)
        buffer_y_pred.seek(0)

        buffer_model = io.BytesIO()
        pickle.dump(model, buffer_model)
        buffer_model.seek(0)
        
        data = {'project_id' : self.pk, 'model_name' : model_name, 'scores': json.dumps(scores), 'threshold' : threshold, 'depth_scores' : json.dumps(depth_scores)}
        files = { 'y_pred_file' : ('_'.join([self.session.username, self.project_name, 'y_pred']),buffer_y_pred), 'model_file' : ('_'.join([self.session.username, self.project_name, model_name]),buffer_model)}
        r = requests.put(self.session.url + '/create_model/' , headers = headers , data = data, files=files)
        if r.status_code != 200 : 
            raise Exception('Request Error  '  + r.text)
        self._models.append(r.json())
        print('Model has been pushed to althiqa\'s platform with the following metric scores :')
        pprint.pprint(scores)


    def compute_metric(self, metric, model,threshold=None):
        
        if type(metric['metric_function']) == str :
            try :
                r = requests.get(metric['metric_function'])
                b= io.BytesIO()
                b.write(r.content)
                b.seek(0)
                m1 = marshal.load(b)
                m = types.FunctionType(m1, globals(), "metric")
            except :
                print(metric)
                raise Exception("Error while loading the metric")
        else : m = metric['metric_function']


        predict  = getattr(model, "predict", None)
        def predict_proba(y):
                return 
        if self.project_type == "classif":
            def predict(y):
                return (pd.DataFrame(model.predict_proba(y))>float(threshold)).iloc[:, 1].astype(int)
            def predict_proba(y):
                return model.predict_proba(y)

        arg_dict = {'model' : model, 'y_pred' : pd.DataFrame(predict(self.x_test)).iloc[:,0].to_numpy(), 'probs' :predict_proba(self.x_test), 'threshold': threshold, 'x_train' : self.x_train, 'y_train' : self.y_train.iloc[:,0].to_numpy(), 'x_test' : self.x_test, 'y_test': self.y_test.iloc[:,0].to_numpy(), 'protected_attribute' : metric.get('protected_attribute',None)}
        f_params = list(inspect.signature(m).parameters.keys())
        d2 = { k:v for k,v in arg_dict.items() if k in f_params }

        if 'protected_attribute' in d2.keys() and d2['protected_attribute'] is None :
            #print("{mn} metric needs a valid protected attribute argument. Metric not computed.".format(mn = metric['metric_name']))
            return None
        
        try :
            score = m(**d2)
        except Exception as e: 
            raise e
        return {'score' : round(float(score),2)}

    def compute_depth_metric(self, metric, model, threshold = None):
        
        predict  = getattr(model, "predict", None)
        if not callable(predict):
            def predict(y):
                return (pd.DataFrame(model.predict_proba(y))>float(threshold)).iloc[:, 1].astype(int)
        y_pred =  pd.DataFrame(predict(self.x_test)).iloc[:,0].to_numpy()
        rank = metric['ranks']
        scores = compute_acc_vs_depth(rank, y_pred, self.y_test.iloc[:,0].to_numpy() )
        return scores

        
    def create_metric(self, metric_name: str, metric_function , protected_attribute = None, description = None, h_is_b = False, v= 1 ):
        # check metric can be applied to y_pred
        try :
            model = MockModel()
            arg_dict = {'model' : model, 'y_pred' : pd.DataFrame(model.predict(self.x_test)).iloc[:,0].to_numpy(), 'x_train' : self.x_train, 'y_train' : self.y_train, 'x_test' : self.x_test, 'y_test': self.y_test, 'protected_attribute' : protected_attribute}
            f_params = list(inspect.signature(metric_function).parameters.keys())
            d2 = { k:v for k,v in arg_dict.items() if k in f_params }
            score = float(metric_function(**d2))
            if not (type(score) == float ) or (type(score) == int ) :
                raise ValueError()
        except Exception as e :
            pass 
            
        metric = {'metric_name' : metric_name, 'metric_type' : self.project_type, 'metric_function' : metric_function, 'description' : description ,'h_is_b' : h_is_b, 'protected_attribute' : protected_attribute}
        scores = {}
        for model in self._models : 
            m = self.get_model(model['model_name'])
            score = self.compute_metric(metric, m, model['threshold'])
            if score is not None : 
                scores[model['model_name']] =  score
        metric['scores'] = scores
        
        headers = {}
        headers["Accept"] = "application/json"
        headers["Authorization"] = "Bearer {token}".format(token  = self.session.token)
        metric_function_file = io.BytesIO()
        marshal.dump(metric['metric_function'].__code__, metric_function_file)
        metric_function_file.seek(0)
        
        data = {'project_id' : self.pk, 'metric_name' : metric['metric_name'], 'metric_type' : metric['metric_type'],
         'description' :  metric['description'] , 'h_is_b' :  metric['h_is_b'], 'scores': json.dumps(scores)}
        files = { 'metric_file' : ('_'.join([self.session.username, self.project_name, metric_name]),metric_function_file)}
        r = requests.put(self.session.url + '/create_metric/' , headers = headers , data = data, files=files)
        if r.status_code != 200 : 
            raise Exception('Request Error  '  + r.text)
        self._metrics.append(metric)
        if v : 
            print('Metric has been pushed to the Althiqa platform with the following metric scores :')
            pprint.pprint(scores)

    def create_depth_metric(self, depth_metric_name: str, depth_metric_function , description = None):
        # check metric can be applied to y_pred
        try :
            model = MockModel()
            arg_dict = {'model' : model, 'y_pred' : model.predict(self.y_test), 'x_train' : self.x_train, 'y_train' : self.y_train, 'x_test' : self.x_test, 'y_test': self.y_test, 'protected_attribute' : protected_attribute}
            f_params = list(inspect.signature(depth_metric_function).parameters.keys())
            d2 = { k:v for k,v in arg_dict.items() if k in f_params }
            score = list(depth_metric_function(**d2))
            rank = get_rank(score)
            if not (type(score) == float ) or (type(score) == int ) :
                raise ValueError()
        except Exception as e : 
            raise e 
            
        depth_metric = {'depth_metric_name' : depth_metric_name,  'depth_metric_function' :depth_metric_function, 'description' : description, 'rank': rank }
        scores = {}
        for model in self._models :
         
            m = self.get_model(model['model_name'])
            score = self.compute_depth_metric(depth_metric, m, model['threshold'])
            scores[model['model_name']] =  score

        depth_metric['scores'] = scores
        
        headers = {}
        headers["Accept"] = "application/json"
        headers["Authorization"] = "Bearer {token}".format(token  = self.session.token)
        metric_function_file = io.BytesIO()
        pickle.dump(depth_metric['metric_function'], metric_function_file)
        metric_function_file = io.BytesIO().seek(0)
        
        data = {'project_id' : self.pk, 'depth_metric_name' : depth_metric['depth_metric_name'],
         'description' :  depth_metric['description'] , 'rank' : depth_metric['rank'], 'scores': json.dumps(scores)}
        files = { 'depth_metric_file' : ('_'.join([self.session.username, self.project_name, depth_metric_name]),metric_function_file)}
        r = requests.put(self.session.url + '/create_depth_metric/' , headers = headers , data = data, files=files)
        if r.status_code != 200 : 
            raise Exception('Request Error  '  + r.text)
        print('Depth metric has been pushed to althiqa\'s platform with the following metric scores :')
        pprint.pprint(scores)

    







class Session():
    def __init__(self, username, password, url ='http://api.althiqa.io') -> None:
        self.url = url
        self.username = username
        self.password = password
        try :
            r = requests.post(self.url + '/api/token/', data = {'username' : self.username , 'password' : self.password} )
            if r.status_code != 200 :
                raise Exception('Request failed, is your url correct ?')
            res = r.json()
            if 'access' in res.keys() :
                print('Login succesful, token obtained')
                self.token = res['access']
            else : 
                raise Exception("Login unsuccessful : "  +  r.text)
        except Exception as e :
            raise Exception("Can't authenticate to 'url' with this username and password")

    def list_projects(self):
        headers = {}
        headers["Accept"] = "application/json"
        headers["Authorization"] = "Bearer {token}".format(token  = self.token)
        r = requests.get(self.url + '/list_projects/', headers = headers)
        if r.status_code != 200 : 
            raise Exception('Bad request : '  + r.text)
        pprint.pprint(r.json())
    
            
    
    def create_project(self, project_name :str, x_train, y_train, x_test, y_test, project_type = 'reg') :
        # check project_type in 'classif' or 'reg'
        if project_type not in ['reg', 'classif'] : 
            raise Exception("Argument project_type must be 'reg' (default) or 'classif'")
        # check datas can be converted to dataframe
        try : 
            x_train = pd.DataFrame(x_train)
            y_train = pd.DataFrame(y_train)
            x_test = pd.DataFrame(x_test)
            y_test = pd.DataFrame(y_test)
        except :
            raise Exception("Passed datas must be convertible to DataFrame")
        headers = {}
        headers["Accept"] = "application/json"
        headers["Authorization"] = "Bearer {token}".format(token  = self.token)
        data = {'project_name' :project_name ,'project_type' : project_type.capitalize()}
        buffer_x_train = io.BytesIO()
        buffer_y_train = io.BytesIO()
        buffer_x_test = io.BytesIO()
        buffer_y_test = io.BytesIO()

        x_train.to_pickle(buffer_x_train)
        y_train.to_pickle(buffer_y_train)
        x_test.to_pickle(buffer_x_test)
        y_test.to_pickle(buffer_y_test)

        buffer_x_train.seek(0)
        buffer_x_train.name = 'x_train.pckl'
        buffer_y_train.seek(0)
        buffer_x_test.seek(0)
        buffer_y_test.seek(0)
        
        files = { 'x_train_file' : ('_'.join([self.username, project_name, 'x_train']),buffer_x_train), 'y_train_file' : ('_'.join([self.username, project_name, 'y_train']),buffer_y_train),
                'x_test_file': ('_'.join([self.username, project_name, 'x_test']),buffer_x_test), 'y_test_file' : ('_'.join([self.username, project_name, 'y_test']),buffer_y_test)}
        
        for key, value in files.items()  :
            nmb = round(value[1].getbuffer().nbytes / (1024*1024),2)
            if nmb > MAX_MB_UPLOAD : 
                raise Exception(f"File {key} is too large ({nmb}MB). Please use objects smaller than {MAX_MB_UPLOAD}MB.".format(key=key, nmb=nmb ))
        
        default_depth_rank = []
        if project_type=='classif':
            default_depth_rank = get_rank(compute_aiirw_depth(x_train, y_train, x_test, y_test) )       
        data = {'project_name' : project_name, 'project_type' : project_type, 'y_len' : len(y_test), 'default_depth_rank' : json.dumps(default_depth_rank.tolist())}
        
        r = requests.put(self.url + '/create_project/' , headers = headers , data = data, files=files)
        if r.status_code != 200 : 
            raise Exception('Request Error  '  + r.text)
        res = r.json()
        pk = res['pk']
        #print('Project pk:' + str(pk))
        #print(res)
        p = Project(self, pk, project_name, project_type, x_train, y_train, x_test, y_test, metrics = [], depth_metrics = res['depth_metrics'])
        for m in d_metrics :
            if m['metric_type'] == p.project_type : p.create_metric(metric_name = m['metric_name'], metric_function= m['metric_function'] , protected_attribute = m.get('protected_attribute',None), description = m['description'], h_is_b= m['h_is_b'], v=0)
        return p

    def get_project(self, name : str) :
        headers = {}
        headers["Accept"] = "application/json"
        headers["Authorization"] = "Bearer {token}".format(token  = self.token)
        params = {'name' : name}
        r = requests.get(self.url + '/get_project' , headers = headers , params = params)
        res = r.json()
        x_train = self.get_s3_file(res['x_train'])
        y_train = self.get_s3_file(res['y_train']) # Faire comme cela ??
        x_test = self.get_s3_file(res['x_test'])
        y_test = self.get_s3_file(res['y_test'])
        pk = res['pk']
        print(res['project_type'])
        return Project(self, pk, res['project_name'], res['project_type'], x_train, y_train, x_test, y_test, models = res['models'] , metrics= res['metrics'] , depth_metrics=res['depth_metrics'] , scores = res['scores'], reports= res['reports'])

    def get_s3_file(self, url):
        # A Tester / Debugger
        print(url)
        r = requests.get(url) # Coder l'API en faisant attention à vérifier que le fichier appartient bien au user.
        buff = r.content
        print(type(buff))
        b= io.BytesIO()
        b.write(r.content)
        b.seek(0)
        df_obj = pickle.load(b)
        return df_obj
